/*
 * kgraft_patch_bsc1013543
 *
 * Fix for CVE-2016-9794, bsc#1013543
 *
 *  Upstream commit:
 *  3aa02cb664c5 ("ALSA: pcm : Call kill_fasync() in stream lock")
 *
 *  SLE12(-SP1) commit:
 *  1a022ee86660d4234e89e3bfb1bd64618dacc9e1
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <sound/core.h>
#include <sound/pcm.h>

#define KGR_PATCHED_MODULE      "snd_pcm"

static int (*kgr_snd_pcm_update_hw_ptr0)(struct snd_pcm_substream *substream, unsigned int in_interrupt);
static void (*kgr_snd_timer_interrupt)(struct snd_timer * timer, unsigned long ticks_left);
static rwlock_t *kgr_snd_pcm_link_rwlock;

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "snd_pcm_update_hw_ptr0", (void *)&kgr_snd_pcm_update_hw_ptr0 },
	{ "snd_timer_interrupt", (void *)&kgr_snd_timer_interrupt },
	{ "snd_pcm_link_rwlock", (void *)&kgr_snd_pcm_link_rwlock },
};

#define kgr_snd_pcm_stream_lock_irqsave(substream, flags) \
do { \
	read_lock_irqsave(kgr_snd_pcm_link_rwlock, (flags)); \
	spin_lock(&substream->self_group.lock); \
} while (0)

#define kgr_snd_pcm_stream_unlock_irqrestore(substream, flags) \
do { \
	spin_unlock(&substream->self_group.lock); \
	read_unlock_irqrestore(kgr_snd_pcm_link_rwlock, (flags)); \
} while (0)

/* patched function */
void kgr_snd_pcm_period_elapsed(struct snd_pcm_substream *substream)
{
	struct snd_pcm_runtime *runtime;
	unsigned long flags;

	if (PCM_RUNTIME_CHECK(substream))
		return;
	runtime = substream->runtime;

	if (runtime->transfer_ack_begin)
		runtime->transfer_ack_begin(substream);

	kgr_snd_pcm_stream_lock_irqsave(substream, flags);
	if (!snd_pcm_running(substream) ||
	    kgr_snd_pcm_update_hw_ptr0(substream, 1) < 0)
		goto _end;

	if (substream->timer_running)
		kgr_snd_timer_interrupt(substream->timer, 1);
 _end:
	/*
	 * Fix CVE-2016-9794
	 *  +1 line
	 */
	kill_fasync(&runtime->fasync, SIGIO, POLL_IN);
	kgr_snd_pcm_stream_unlock_irqrestore(substream, flags);
	if (runtime->transfer_ack_end)
		runtime->transfer_ack_end(substream);
	/*
	 * Fix CVE-2016-9794
	 *  -kill_fasync(&runtime->fasync, SIGIO, POLL_IN);
	 */
}

static int kgr_patch_bsc1013543_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1013543_module_notify(struct notifier_block *nb,
	unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1013543_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1013543_module_nb = {
	.notifier_call = kgr_patch_bsc1013543_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1013543_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1013543_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1013543_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1013543_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1013543_module_nb);
}
